@@ -5,8 +5,6 @@
 
 namespace xs {
 
-using std::invalid_argument;
-
 my_perl_auto_t my_perl;
 
 payload_marker_t sv_payload_default_marker;
@@ -34,7 +32,7 @@ static SV* _next_create_wrapper (pTHX_ CV* cv, next_t type) {
         case NEXT_METHOD: next_code = "next::method"; break;
         case NEXT_MAYBE:  next_code = "maybe::next::method"; break;
     }
-    if (!next_code.length()) throw new invalid_argument("type");
+    if (!next_code.length()) throw std::invalid_argument("type");
     std::string code =
         "package " + stashname + ";\n" +
         "use feature 'state';\n" +
@@ -88,42 +86,55 @@ SV* call_next (pTHX_ CV* cv, SV** args, I32 items, next_t type, I32 flags) {
     return ret;
 }
 
-static inline HV* _get_stash (pTHX_ HV* stash)         { return stash; }
-static inline HV* _get_stash (pTHX_ SV* CLASS)         { return gv_stashsv(CLASS, GV_ADD); }
-static inline HV* _get_stash (pTHX_ const char* CLASS) { return gv_stashpvn(CLASS, strlen(CLASS), GV_ADD); }
+/* should be called when interpreter is cloned. If we get here then our perl_object is cloned, that is it is present in PL_ptr_table
+ * So we only need to get the new pointer instead of the old one */
+void XSBackref::on_perl_dup (pTHX_ int32_t refcnt) {
+    if (perl_object) {
+        perl_object = MUTABLE_SV(ptr_table_fetch(PL_ptr_table, perl_object));
+        assert(perl_object); assert(refcnt);
+        SvREFCNT(perl_object) = refcnt;
+    }
+}
 
-template <typename C>
-static inline SV* _typemap_out_oext_ (pTHX_ SV* obase, void* var, C CLASS, payload_marker_t* marker) {
-    if (!var) return &PL_sv_undef;
-    SV* objrv;
-    if (obase) {
-        if (SvROK(obase)) {
-            objrv = obase;
-            obase = SvRV(obase);
-        }
-        else {
+namespace _tm {
+
+    static inline HV* _get_stash (pTHX_ HV* stash)         { return stash; }
+    static inline HV* _get_stash (pTHX_ SV* CLASS)         { return gv_stashsv(CLASS, GV_ADD); }
+    static inline HV* _get_stash (pTHX_ const char* CLASS) { return gv_stashpvn(CLASS, strlen(CLASS), GV_ADD); }
+
+    template <typename C>
+    static inline SV* _out_oext_ (pTHX_ SV* obase, void* var, C CLASS, payload_marker_t* marker) {
+        if (!var) return &PL_sv_undef;
+        SV* objrv;
+        if (obase) {
+            if (SvROK(obase)) {
+                objrv = obase;
+                obase = SvRV(obase);
+            }
+            else {
+                objrv = newRV_noinc(obase);
+                sv_bless(objrv, _get_stash(aTHX_ CLASS));
+            }
+        } else {
+            obase = newSV(0);
             objrv = newRV_noinc(obase);
             sv_bless(objrv, _get_stash(aTHX_ CLASS));
         }
-    } else {
-        obase = newSV(0);
-        objrv = newRV_noinc(obase);
-        sv_bless(objrv, _get_stash(aTHX_ CLASS));
+        sv_payload_attach(aTHX_ obase, var, marker);
+        return objrv;
     }
-    sv_payload_attach(aTHX_ obase, var, marker);
-    return objrv;
-}
 
-SV* _typemap_out_oext (pTHX_ SV* obase, void* var, HV* CLASS, payload_marker_t* marker) {
-    return _typemap_out_oext_(aTHX_ obase, var, CLASS, marker);
-}
+    SV* _out_oext (pTHX_ SV* obase, void* var, HV* CLASS, payload_marker_t* marker) {
+        return _out_oext_(aTHX_ obase, var, CLASS, marker);
+    }
 
-SV* _typemap_out_oext (pTHX_ SV* obase, void* var, SV* CLASS, payload_marker_t* marker) {
-    return _typemap_out_oext_(aTHX_ obase, var, CLASS, marker);
-}
+    SV* _out_oext (pTHX_ SV* obase, void* var, SV* CLASS, payload_marker_t* marker) {
+        return _out_oext_(aTHX_ obase, var, CLASS, marker);
+    }
 
-SV* _typemap_out_oext (pTHX_ SV* obase, void* var, const char* CLASS, payload_marker_t* marker) {
-    return _typemap_out_oext_(aTHX_ obase, var, CLASS, marker);
+    SV* _out_oext (pTHX_ SV* obase, void* var, const char* CLASS, payload_marker_t* marker) {
+        return _out_oext_(aTHX_ obase, var, CLASS, marker);
+    }
 }
 
 };
